#include <cell/pad.h>
#include <cell/rtc.h>
#include <cell/cell_fs.h>
#include <cell/sysmodule.h>
#include <sysutil/sysutil_sysparam.h>
#include <sysutil/sysutil_music2.h>
#include <sysutil/sysutil_common.h>
#include <sysutil/sysutil_search.h>
#include <sysutil/sysutil_rec.h>

#include "inc/exports.h"
#include "misc.h"
#include "mem.h"
#include "blitting.h"
#include "pad_hook.h"
#include "game_menu.h"

#ifdef DEBUG
#include "network.h"
#endif

#ifdef HAVE_SSCROLLER
#include "scroller.h"
#endif

#ifdef HAVE_STARFIELD
#include "starfield.h"
#endif

// game menu specific methods
static int32_t load_modules(void);
static int32_t unload_modules(void);
static int32_t start_rec(void);
static int32_t stop_rec(void);
static int32_t getTrackInfo(void);
static void cb_recutil(int32_t cb_status, int32_t cb_error, void *cb_userdata);
static void cb_musicutil(uint32_t cb_event, void *cb_param, void *cb_userdata);
static void cb_searchutil(CellSearchEvent cb_event, int32_t cb_result, const void *cb_param, void *cb_userdata);

// game menu data
static uint8_t plugin_running;  // pluign running
CellPadData mypdata;            // also used in pad_hook.c

// music util data
enum {MIN_VOL = 0, DEF_VOL = 5, MAX_VOL = 9};
static int32_t currentVolume = DEF_VOL;
static char currentPlayStatus[10], currentTrack[CELL_SEARCH_TITLE_LEN_MAX+1];
static int16_t volumeValue[10] = { 0.0f /* off */, 0.39811f/* -8db */, 0.50119f, 0.63096f/* -4db */, 0.79433f, 1.0f/* 0 db */, 1.25893f, 1.58489f/* 4db */, 1.99526f, 2.51189f/* 8db */};

// recording util data
static sys_memory_container_t rec_container = SYS_MEMORY_CONTAINER_ID_INVALID;

////////////////////////////////////////////////////////////////////////
// GAME BLITTING
bool menu_running = 0;          // menu off(0) or on(1)
static int8_t line = 0;         // current line into menu, init 0 (entry 1:)
static int8_t view = 0;         // menu view, init 0 (main view)
#define MAX_MENU_LINE 11
#define MAX_MENU_VIEW 4
enum view_title {MAIN_VIEW, MUSIC_VIEW, SCREENSHOT_VIEW, RECORDING_VIEW};
static int8_t view_menu[MAX_MENU_VIEW] = {4, 11, 3, 3};
const char *entry_str[MAX_MENU_VIEW][MAX_MENU_LINE] = {
{
	"Open music menu",
	"Open screenshot menu",
	"Open recording menu",
	"Unload plugin",
},
{   // music menu
  "Back to main view",
	"SELECT TRACK",
  "PLAY    |>",
  "PAUSE   ||",
	"STOP    []",
	"VOL      +",
	"VOL      -",
  "NEXT    >|",
	"PREV    |<",
  "FFWD    >>",
	"RWD     <<",
},
{ // screenshot menu
  "Back to main view",
  "screenshot(Game)",
  "screenshot(Game+Menu)",
},
{ // recording menu
  "Back to main view",
  "start recording",
  "stop recording",
}
};

/***********************************************************************
* draw a frame (in game)
***********************************************************************/
static void draw_frame(void)
{
	int32_t i;
  char tempstr[128];

	// all 32bit colors are ARGB, the framebuffer format
	set_background_color(0xFF000000);     // black, opaque
	set_foreground_color(0xFFFFFFFF);     // white, opaque

	// fill background with background color
	draw_background();

  #ifdef HAVE_STARFIELD
  draw_stars();       // to keep them under text lines
  #endif

  #ifdef HAVE_PNG_BG
	// draw background from png
  draw_png(0, 0, 0, 0, 0, 720, 400);
  #endif

	// draw menu entry list
	for(i = 0; i < view_menu[view]; i++)
  {
    i == line ? set_foreground_color(0xFFFF0000) : set_foreground_color(0xFFFFFFFF);
    print_text(10, 10 + (FONT_H * (i + 1)), entry_str[view][i]);
  }

  #ifdef HAVE_SSCROLLER
  // testing sine in a scroller
  draw_text(330);
  move_text();
  #endif

  // menu specific printing
  set_foreground_color(0xFFFFFFFF);
  if (view == MAIN_VIEW) {
	  print_text(300, 8, "Main Menu");
  } else if (view == MUSIC_VIEW) {
	  print_text(300, 8, "Music Menu");
    sprintf(tempstr, "Track: %s", currentTrack);
    print_text(200, 150 + (FONT_H * (0)), tempstr);
    sprintf(tempstr, "Volume: %d", currentVolume);
    print_text(200, 150 + (FONT_H * (1)), tempstr);
    sprintf(tempstr, "Playback Status: %s", currentPlayStatus);
    print_text(200, 150 + (FONT_H * (2)), tempstr);
    set_foreground_color(0xFF00FF00);
    sprintf(tempstr, "Note: You may also select a track by holding down [Start] and pressing\n[Select+R2]");
    print_text(10, 10 + (FONT_H * (i + 2)), tempstr);
  } else if (view == SCREENSHOT_VIEW) {
	  print_text(300, 8, "Screenshot Menu");
    set_foreground_color(0xFF00FF00);
    sprintf(tempstr, "Note: You may also take a screenshot by holding down [Start] and pressing\n[Select+R1]");
    print_text(10, 10 + (FONT_H * (i + 2)), tempstr);
  } else if (view == RECORDING_VIEW) {
	  print_text(300, 8, "Recording Menu");
    set_foreground_color(0xFF00FF00);
    sprintf(tempstr, "Note: You may also start/stop recording by holding down [Start] and pressing\n[Select+L1] to start recording and\n[Select+L2] to stop recording");
    print_text(10, 10 + (FONT_H * (i + 2)), tempstr);
  }
}

/***********************************************************************
* stop game menu
***********************************************************************/
static void stop_GAME_Menu(void)
{
  menu_running = 0;     // menu off
  #ifdef HAVE_SYS_FONT
  font_finalize();      // unbind renderer and kill font-instance
  #endif
  destroy_heap();       // free heap memory
  rsx_fifo_pause(1, 0); // continue rsx rendering
}

/***********************************************************************
* start game menu
***********************************************************************/
static int32_t start_GAME_Menu(void)
{
  int32_t ret, mem_size;

  line = 0;                                                         // menu on first entry in list
  rsx_fifo_pause(1, 1);
  mem_size =(((CANVAS_W * CANVAS_H * sizeof(uint32_t)) * 2)/*fg & bg*/ + MB(2)/*leave some head room*/) / MB(1)/*convert to multiple of page size 1MB*/;
  ret = create_heap(mem_size);
  if (ret != 0) {
    return(-1);
  }
  ret = init_graphic();                                             // initialize game menu graphic
  if (ret != 0) {
    return(-1);
  }
  #ifdef HAVE_PNG_BG
  ret = load_png_bitmap(0, PNG_BG_PATH);                            // load a png background if you like
  if (ret != 0) {
    return(-1);
  }
  #endif
  menu_running = 1;                                                 // set menu_running on
  return(0);
}

/***********************************************************************
* execute a menu action, based on line(current selected menu entry)
***********************************************************************/
static void do_menu_action(void)
{
  switch(view)
  {
    case 0:                     // "Main Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Open Music menu"
          view = 1;
          line = 0;
          break;
        case 1:                 // "Menu Entry 2: Open screenshot menu"
          view = 2;
          line = 0;
          break;
        case 2:                 // "Menu Entry 3: Open recording menu"
          view = 3;
          line = 0;
          break;
        case 3:                 // "Menu Entry 4: Unload plugin"
          stop_GAME_Menu();
          remove_pad_hook();
          plugin_running = 0;
          break;
      }
      break;
    case 1:                     // "Music Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Back to main menu"
          view = line = 0;
          break;
        case 1:                 // "Menu Entry 2: SELECT TRACK"
          stop_GAME_Menu();
          cellMusicSelectContents2();
          break;
        case 2:                 // "Menu Entry 3: PLAY"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_PLAY, NULL);
          sprintf(currentPlayStatus, "PLAYING");
          break;
        case 3:                 // "Menu Entry 4: PAUSE"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_PAUSE, NULL);
          sprintf(currentPlayStatus, "PAUSE");
          break;
        case 4:                 // "Menu Entry 5: STOP"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_STOP, NULL);
          sprintf(currentPlayStatus, "STOPPED");
          break;
        case 5:                 // "Menu Entry 6: VOL+"
          currentVolume = (currentVolume > MAX_VOL) ? currentVolume : currentVolume + 1;
          cellMusicSetVolume2(volumeValue[currentVolume]);
          break;
        case 6:                 // "Menu Entry 7: VOL-"
          currentVolume = (currentVolume < MIN_VOL) ? currentVolume : currentVolume - 1;
          cellMusicSetVolume2(volumeValue[currentVolume]);
          break;
        case 7:                 // "Menu Entry 8: NEXT"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_NEXT, NULL);
          break;
        case 8:                 // "Menu Entry 9: PREV"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_PREV, NULL);
          break;
        case 9:                 // "Menu Entry 10: FFWD"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_FASTFORWARD, NULL);
          break;
        case 10:                // "Menu Entry 11: RWD"
          cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_FASTREVERSE, NULL);
          break;
      }
      break;
    case 2:                     // "Screenshot Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Back to main menu"
          view = line = 0;
          break;
        case 1:                 // "Menu Entry 2: screenshot(Game)"
          screenshot(0);
          stop_GAME_Menu();
          break;
        case 2:                 // "Menu Entry 3: screenshot(Game+Menu)"
          screenshot(1);
          stop_GAME_Menu();
          break;
      }
      break;
    case 3:                     // "Recording Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Back to main menu"
          view = line = 0;
          break;
        case 1:                 // "Menu Entry 2: start recording"
          start_rec();
          stop_GAME_Menu();
          break;
        case 2:                 // "Menu Entry 3: stop recording"
          stop_rec();
          stop_GAME_Menu();
          break;
      }
      break;
  }
}

/***********************************************************************
* game menu main
***********************************************************************/
int32_t game_menu(void)
{
  #ifdef DEBUG
	dbg_init();
	dbg_printf("programstart:\n");
  #endif
	uint32_t oldpad = 0, curpad = 0;
  int32_t ret = load_modules();
  if (ret < 0) {
    return(ret);
  }

  #ifdef HAVE_STARFIELD
  init_once(/* stars */);
  #endif

  // set up hook for interceptin pad data
  setup_pad_hook();
  plugin_running = 1;
	while(plugin_running)
	{
		if(!menu_running)                                                     // GAME menu is not running, normal game execution
		{
      if(mypdata.len > 0)                                                 // if pad data
			{
				curpad = (mypdata.button[2] | (mypdata.button[3] << 8));          // merge pad data
        memset(&mypdata, 0, sizeof(CellPadData));
        if (curpad != oldpad)
        {
          if(curpad & PAD_SELECT)
          {
            if(curpad & PAD_START)
            {
              if(curpad & PAD_R1)                                         // if select+start+R1 was pressed, take screenshot
              {
                screenshot(0);
              }
              else if(curpad & PAD_R2)                                    // if select+start+L1 was pressed, start recording
              {
                cellMusicSelectContents2();
              }
              else if(curpad & PAD_L1)                                    // if select+start+L2 was pressed, stop recording
              {
                start_rec();
              }
              else if(curpad & PAD_L2)                                    // if select+start+R2 was pressed, select a track
              {
                stop_rec();
              }
            }
            else if(!(curpad & PAD_START))                                // if select was pressed, start GAME menu
            {
              ret = start_GAME_Menu();
              if (ret != 0)
              {
                stop_GAME_Menu();                                         // failed to init graphics, try pressing select again
              }
            }
          }
        }
        oldpad = curpad;
			}
      else
      {
				oldpad = 0;
      }
			sys_timer_usleep(70000);                                          // game sync
		}
		else                                                                // menu is running, draw frame, flip frame, check for pad events, sleep, ...
		{
			draw_frame();
		  flip_frame();
      uint32_t type;
      cellPadGetDataExtra(0, &type, &mypdata);                          // use alternate get pad method
			if(mypdata.len > 0)                                               // if pad data
			{
				curpad = (mypdata.button[2] | (mypdata.button[3] << 8));        // merge pad data
        memset(&mypdata, 0, sizeof(CellPadData));
        if (curpad != oldpad)
        {
          if(curpad & PAD_SELECT)
          {
            if(curpad & PAD_START)
            {
              if(curpad & PAD_R1)                                       // if select+start+R1 was pressed, take screenshot
              {
                screenshot(1);
              }
            }
            else if (!(curpad & PAD_START))                             // if select was pressed stop VSH menu
            {
              stop_GAME_Menu();
              sys_timer_usleep(100000);
            }
          }
          else if(curpad & PAD_UP)
          {
            if(line > 0)
            {
              line--;
            }
            else
            {
              line = 0;
            }
          }
          else if(curpad & PAD_DOWN)
          {
            if(line < (view_menu[view]-1))
            {
              line++;
            }
            else
            {
              line = (view_menu[view]-1);
            }
          }
          else if(curpad & PAD_CROSS)
          {
            do_menu_action();
          }
        }
        oldpad = curpad;
      }
			else
      {
				oldpad = 0;
      }
			sys_timer_usleep(200);                                              // short menu frame delay
		}
	}

  // unloading plugin...
  unload_modules();
  #ifdef DEBUG
	dbg_fini();
  #endif
  return(0);
}

////////////////////////////////////////////////////////////////////////
// GAME PLUGIN SPECIFIC METHODS
/***********************************************************************
* load system modules
***********************************************************************/
static int32_t load_modules(void)
{
  int32_t ret;

  // load necessary png and font modules for blitting
  ret = cellSysmoduleLoadModule(CELL_SYSMODULE_PNGDEC);
  if (ret != CELL_OK) {
    return(ret);
  }
  ret = cellSysmoduleLoadModule(CELL_SYSMODULE_FONT);
  if (ret != CELL_OK) {
    return(ret);
  }

  // module for setting screenshot date
  ret = cellSysmoduleLoadModule(CELL_SYSMODULE_RTC);
  if (ret != CELL_OK) {
    return(ret);
  }

  // modules for in game music
	ret = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_SEARCH);
	if (ret != CELL_OK ) {
    return(ret);
	}
	ret = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_MUSIC2);
	if (ret != CELL_OK ) {
    return(ret);
	}

  // initialize search and music modules
  ret = cellSearchInitialize( CELL_SEARCH_MODE_NORMAL, SYS_MEMORY_CONTAINER_ID_INVALID, cb_searchutil, NULL);
	if (ret != CELL_MUSIC2_OK) {
    return(ret);
  }
	ret = cellMusicInitialize2(CELL_MUSIC2_PLAYER_MODE_NORMAL, 250, cb_musicutil, NULL);
	if (ret != CELL_MUSIC2_OK) {
    return(ret);
  }

  // initialize recording module
  ret = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_REC);
  if (ret != CELL_OK) {
	  return(ret);
  }
  return(0);
}

/***********************************************************************
* unload system modules
***********************************************************************/
static int32_t unload_modules(void)
{
  int32_t ret;

  /*ret = cellMusicFinalize2();
  if ( ret != CELL_OK ) {
    return(ret);
  }*/
	ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_MUSIC2);
	if (ret != CELL_OK ) {
	  return(ret);
  }
	ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_SEARCH);
	if (ret != CELL_OK ) {
	  return(ret);
  }
  ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_REC);
  if (ret != CELL_OK) {
	  return(ret);
  }
  ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_VIDEO_EXPORT);
  if (ret != CELL_OK) {
	  return(ret);
  }
  ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_GAME);
  if (ret != CELL_OK) {
	  return(ret);
  }

  // leave some modules loaded in case they are used by the game
  /*ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_RTC);
  if (ret != CELL_OK) {
    return(ret);
  }
  ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_FONT);
  if (ret != CELL_OK) {
    return(ret);
  }
  ret = cellSysmoduleUnloadModule(CELL_SYSMODULE_PNGDEC);
  if (ret != CELL_OK) {
    return(ret);
  }*/
  return(0);
}

/***********************************************************************
* setup and start recording
***********************************************************************/
static int32_t start_rec(void)
{
  CellRecParam recParam;
  CellRecOption recOpt[15];
  int32_t ret, index = 0;
  uint32_t fsBlockSize;
  uint64_t fsFreeBlockCount;
  size_t memSize;

  // check if there's enough free space on hdd
  ret = cellFsGetFreeSize("/dev_hdd0/", &fsBlockSize, &fsFreeBlockCount);
  if (ret < 0 ) {
    return(ret);
  }
  if (fsFreeBlockCount < 1 ) {
    return(-1);
  }

  // create recording directory
  /*ret = cellFsMkdir("/dev_hdd0/recordings/", (CELL_FS_S_IFDIR | CELL_FS_S_IRWXU));
  if (ret < 0 && ret != EEXIST) {
    return(-1);
  }*/

  // set recording paramters and options
  memset (&recParam, 0, sizeof(recParam));
  recParam.videoFmt = CELL_REC_PARAM_VIDEO_FMT_M4HD_HD720_5000K_30FPS;
  recParam.audioFmt = CELL_REC_PARAM_AUDIO_FMT_AAC_128K;
  recOpt[index].option = CELL_REC_OPTION_XMB_BGM;
  recOpt[index++].value.xmb_bgm = CELL_REC_PARAM_XMB_BGM_ENABLE;
  recOpt[index].option = CELL_REC_OPTION_SHOW_XMB;
  recOpt[index++].value.show_xmb = 1;
  recOpt[index].option = CELL_REC_OPTION_MPEG4_FAST_ENCODE;
  recOpt[index++].value.mpeg4_fast_encode = 1;
  recParam.numOfOpt = index;
  recParam.pOpt = recOpt;

  // query & allocate memory
  memSize = cellRecQueryMemSize(&recParam);
  ret = sys_memory_container_create (&rec_container, memSize);
  if (ret < 0) {
    cellRecClose(1); // discard file
    return(-1);
  }

  // build file name
  CellRtcDateTime t;
  cellRtcGetCurrentClockLocalTime(&t);
  char fname[128];
  sprintf(fname, "recording_%02d_%02d_%02d_%02d_%02d_%02d.mp4", t.year, t.month, t.day, t.hour, t.minute, t.second);

  // after open, the callback will start the recording
  ret = cellRecOpen("/dev_hdd0/tmp/", fname, &recParam, rec_container, cb_recutil, NULL);
  if (ret < 0) {
    cellRecClose(1);
    return(ret);
  }
  return(0);
}

/***********************************************************************
* stop and close recording
***********************************************************************/
static int32_t stop_rec(void)
{
  int32_t ret;

  // after stopping, the callback will close and export the video
  ret = cellRecStop();
  if (ret < 0) {
    cellRecClose(1);
    return(-1);
  }
  return(0); 
}

/***********************************************************************
* get track title
***********************************************************************/
static int32_t getTrackInfo(void)
{
  CellSearchContentId contents_id;
  CellSearchContentType content_type;

  cellMusicGetContentsId2(&contents_id);
  uint8_t *infoBuffer = (uint8_t *)malloc(CELL_SEARCH_CONTENT_BUFFER_SIZE_MAX);
  if (!infoBuffer) {
    return(-1);
  }
  cellSearchGetContentInfoByContentId(&contents_id, infoBuffer, &content_type);
  CellSearchMusicInfo *musicInfo = (CellSearchMusicInfo *)infoBuffer;
  sprintf(currentTrack, musicInfo->title);
  free(infoBuffer);
  return(0);
}

/***********************************************************************
* recording callback
***********************************************************************/
static void cb_recutil(int32_t cb_status, int32_t cb_error, void *cb_userdata)
{
  int32_t ret;

  if (cb_status == CELL_REC_STATUS_OPEN) {
    ret = cellRecStart();
    if (ret < 0) {
      cellRecClose(1);
    }
  } else if (cb_status == CELL_REC_STATUS_STOP) {
    ret = cellRecClose(0); // close and save
    if (ret < 0) {
      cellRecClose(1);
    }
    if (rec_container != SYS_MEMORY_CONTAINER_ID_INVALID) {
      ret = sys_memory_container_destroy(rec_container);
      if (ret == CELL_OK) {
        rec_container = SYS_MEMORY_CONTAINER_ID_INVALID;
      }
    }
  }
}

/***********************************************************************
* playback callback
***********************************************************************/
static void cb_musicutil(uint32_t cb_event, void *cb_param, void *cb_userData)
{
  if (cb_event == CELL_MUSIC2_EVENT_SELECT_CONTENTS_RESULT) {
    if( (uint32_t)cb_param == CELL_MUSIC2_OK ) {

      // play track immediately after selection
	    cellMusicSetPlaybackCommand2(CELL_MUSIC2_PB_CMD_PLAY, NULL);
      getTrackInfo();
      sprintf(currentPlayStatus, "PLAYING");
    }
  } else if (cb_event ==  CELL_MUSIC2_EVENT_SET_PLAYBACK_COMMAND_RESULT) {
    if ((int32_t)cb_param == CELL_MUSIC2_OK) {
      getTrackInfo();
    }
  }
}

/***********************************************************************
* track search callback
***********************************************************************/
static void cb_searchutil(CellSearchEvent cb_event, int32_t cb_result, const void *cb_param, void *cb_userdata)
{
  // do nothing
}

