#include <sys/syscall.h>
#include <cell/pad.h>
#include <cell/cell_fs.h>

#include "inc/exports.h"
#include "pad_hook.h"

extern CellPadData mypdata; // from game_menu.c
extern bool menu_running; // from game_menu.c

/***********************************************************************
* get process id using syscall
***********************************************************************/
static inline sys_pid_t get_process_id(void)
{
  system_call_0(1);
  return((sys_pid_t)p1);
}

/***********************************************************************
* based off of vsh getNIDfunc but instead of using toc from vsh memory,
* compute NID based on toc of sprx file
*
* const char *sprx_file  = path to sprx file
* const char *prx_name   = name of prx library
* const char *module     = name of module
* uint32_t fnid          = function nid to compute for
* int32_t offset         = offset from nid if needed
***********************************************************************/
static void *getNIDfunc2(const char *sprx_file, const char *prx_name, const char *module, uint32_t fnid, int32_t offset)
{
  char filename[256], fbuf[128];
  uint32_t mod_base, hdr_size, off, phoff, loff;
  int32_t fd, r;
  uint64_t nread;
  uint64_t nbytes;

  sys_prx_id_t prx_id = sys_prx_get_module_id_by_name(prx_name, 0, NULL);
  sys_prx_module_info_t pInfo;
  sys_prx_segment_info_t segment;
  pInfo.filename = &filename[0];
  pInfo.segments = &segment;
  pInfo.filename_size = 256;
  pInfo.segments_num = 1;
  pInfo.size = sizeof(sys_prx_module_info_t);
  r = sys_prx_get_module_info(prx_id, 0, &pInfo);
  mod_base = pInfo.segments->base;

  // read program header offset
  cellFsOpen(sprx_file, CELL_FS_O_RDONLY, &fd, NULL, 0);
  nbytes = 128;
  cellFsReadWithOffset(fd, 0x90/*skip sce header*/, &fbuf[0], nbytes, &nread);

  hdr_size = *(uint32_t *)(fbuf + 0x24);
  off = *(uint32_t *)(fbuf + hdr_size + 0x0C);
  phoff = *(uint32_t *)(fbuf + hdr_size + 0x1C);
  loff = phoff - off; // logical offset

  uint32_t export_start = *(uint32_t *)(mod_base + loff + 0x24);
  uint8_t export_stru_size = (uint8_t)(*(uint32_t *)export_start >> 24);
  while (export_stru_size == 0x1C || export_stru_size == 0x28) {
    uint32_t *export_stru_ptr = (uint32_t *)export_start;
    const char *lib_name_ptr =  (const char *)(*(uint32_t *)((uint8_t *)export_stru_ptr + 0x10));
    if (lib_name_ptr) {
      if(strncmp(module, lib_name_ptr, strlen(lib_name_ptr)) == 0) {
        // we got the proper export struct
        uint32_t lib_fnid_ptr = *(uint32_t *)((char *)export_stru_ptr + 0x14);
        uint32_t lib_func_ptr = *(uint32_t *)((char *)export_stru_ptr + 0x18);
        uint16_t count = *(uint16_t *)((char *)export_stru_ptr + 6); // number of exports
        for (int i = 0; i < count; i++) {
          if (fnid == *(uint32_t *)((char *)lib_fnid_ptr + i*4)) {
            // take address from OPD
            return((void **)*((uint32_t *)(lib_func_ptr) + i) + offset);
          }
        }
      }
    }
    export_stru_size = (uint8_t)(*(uint32_t *)export_start >> 24);
    export_start += export_stru_size;
  }
  return(NULL);
}

/***********************************************************************
* write to process memory using syscall 
***********************************************************************/
static int32_t write_process(uint64_t ea, uint32_t *data, uint32_t size, sys_pid_t pid) {
  if (pid == 0) {
    pid = get_process_id();
  }
  system_call_4(905, (uint64_t)pid, (uint64_t)ea, (uint64_t)size, (uint64_t)((uint32_t)data));
  return_to_user_prog(int32_t);
}

/***********************************************************************
* write hook stup to process memory 
***********************************************************************/
static void write_stub_hook(uint32_t address, uint32_t function, sys_pid_t pid) {
	uint32_t opcode[2];
	opcode[0] = 0x3D600000 + ((function >> 16) & 0xFFFF);  // lis r11, function@high
	opcode[1] = 0x616B0000 + (function & 0xFFFF);  // ori r11, r11, function@low
  write_process((uint64_t)address, (uint32_t *)&opcode, sizeof(uint32_t) * 2, pid);
}

/***********************************************************************
* insert branch to hook function in process memory 
***********************************************************************/
static void branch_ctr(uint32_t address, uint32_t function, sys_pid_t pid) {
  uint32_t opcode[4];
  opcode[0] = 0x3D600000 + ((function >> 16) & 0xFFFF);  // lis r11, function@high
  opcode[1] = 0x616B0000 + (function & 0xFFFF);  // ori r11, r11, function@low
  opcode[2] = 0x7D6903A6; // mtctr r11
  opcode[3] = 0x4E800420; // bctrl
  write_process((uint64_t)address, (uint32_t *)&opcode, sizeof(uint32_t) * 4, pid);
}

/***********************************************************************
* restore this process's toc in hook function
***********************************************************************/
static void restore_toc(uint32_t address, uint32_t toc, sys_pid_t pid) {
	uint32_t opcode[2];
	opcode[0] = 0x3C400000 + ((toc >> 16) & 0xFFFF);  // lis r2, toc@high
	opcode[1] = 0x60420000 + (toc & 0xFFFF);  // ori r2, r2, toc@low
  write_process((uint64_t)address, (uint32_t *)&opcode, sizeof(uint32_t) * 2, pid);
}

/***********************************************************************
* obtain toc from this process's memory 
***********************************************************************/
static uint64_t get_toc(void) {
  uint64_t r2; // r2 == toc
  __asm__ __volatile__(
    "mr %[r2], 2;"
    : [r2] "=&b" (r2) // output
    : // no input
    : "memory"
  );
  return(r2);
}

/***********************************************************************
* main hook to intercept getPadData method. it will use the alternative
* getPadDataExtra method when ps3 menu is not running
***********************************************************************/
void pad_hook(void) {
  uint64_t r3, r4;
  uint32_t port, type;
  CellPadData *pdata;

  // intercept cellPadGetData
  __asm__ __volatile__(
    "mr %[r3], 3;" // pad handle arg
    "mr %[r4], 4;" // pad data arg
    : [r3] "=&r" (r3), [r4] "=&r" (r4)
    :
    : "r0", "r2", "r3", "r4", "r5", "r6", "r7", "r8",
      "r9", "r10", "r11", "r12", "lr", "ctr", "xer",
      "cr0", "cr1", "cr5", "cr6", "cr7", "memory"
  );
  port = (uint32_t)r3;
  pdata = (CellPadData *)(uint32_t)r4;
  cellPadGetDataExtra(port, &type, pdata);
  memcpy(&mypdata, pdata, sizeof(CellPadData));
  if (menu_running) {
    memset(pdata, 0, sizeof(CellPadData));
  }
}

/***********************************************************************
* sets up the hook
***********************************************************************/
int32_t setup_pad_hook(void) {
  uint32_t call_addr, opcode;
  uint32_t toc_offset = sizeof(uint32_t) * 3;
  uint32_t toc_addr = get_toc();

  // disables receiving of pad data and returns the game an untouched pad
  // only menu will receive pad data
  sys_pid_t pid = get_process_id();
  uint32_t hook_stub_addr = (uint32_t)&pad_hook_stub;
  uint32_t hook_function_addr = (uint32_t)&pad_hook_function;
  int32_t (*_cellPadGetData)(uint32_t port_no, CellPadData *data);
  _cellPadGetData = (void *)((int32_t*)getNIDfunc2("/dev_flash/sys/external/libio.sprx", "sys_io_library", "sys_io", 0x8b72cda1, 0));
  if (!(_cellPadGetData)) {
    return(-1);
  }
  call_addr = **(uint32_t **)&_cellPadGetData;

  // copy the first 4 opcodes of the orignal call (cellPadGetData) to the hook's stub
  write_process(hook_stub_addr + (sizeof(uint32_t) * 8), (uint32_t *)call_addr, sizeof(uint32_t) * 4, pid);

  // copy the address of the orignal call (cellPadGetData) at the inststruction where we will resume from
  write_stub_hook(hook_stub_addr + (sizeof(uint32_t) * 12), call_addr + 0x68/*skips getting of pad data from system*/, pid);

  // restore toc for hook function
  restore_toc(hook_function_addr + toc_offset, toc_addr, pid);

  // set return value to CELL_PAD_OK
  opcode = 0x38600000;  // li r3, 0
  write_process(call_addr + 0x68, (uint32_t *)&opcode, sizeof(uint32_t), pid); // replace nop

  // replace the first 4 opcodes of the orignal call (cellPadGetData) to call the hook_stub instead, the hook stub will call the hook function
  branch_ctr(call_addr, hook_stub_addr, pid);

  return(CELL_OK);
}

/***********************************************************************
* removes the hook
***********************************************************************/
int32_t remove_pad_hook(void) {
  uint32_t call_addr, opcode;

  // undo hooking
  sys_pid_t pid = get_process_id();
  uint32_t hook_stub_addr = (uint32_t)&pad_hook_stub;
  int32_t (*_cellPadGetData)(uint32_t port_no, CellPadData *data);
  _cellPadGetData = (void *)((int32_t*)getNIDfunc2("/dev_flash/sys/external/libio.sprx", "sys_io_library", "sys_io", 0x8b72cda1, 0));
  if (!(_cellPadGetData)) {
    return(-1);
  }
  call_addr = **(uint32_t **)&_cellPadGetData;
  write_process((uint64_t)call_addr, (uint32_t *)(hook_stub_addr + (sizeof(uint32_t) * 8)), sizeof(uint32_t) * 4, pid);
  opcode = 0x60000000;  // nop
  write_process(call_addr + 0x68, (uint32_t *)&opcode, sizeof(uint32_t), pid);
  return(CELL_OK);
}

